import pandas as pd
import numpy as np
from pathos.multiprocessing import ProcessPool
import multiprocessing
import inspect
from scipy.stats import rankdata
from tuneta.config import *
from tuneta.optimize import Optimize
import pandas_ta as pta
from finta import TA as fta
import talib as tta
import re
from tabulate import tabulate
from tuneta.optimize import col_name
from collections import OrderedDict


class TuneTA():

    def __init__(self, n_jobs=multiprocessing.cpu_count() - 1, verbose=False):
        self.fitted = []
        self.n_jobs = n_jobs
        self.verbose = verbose

    def fit(self, X, y, trials=5, indicators=['tta'], ranges=[(3, 180)],
        spearman=True, weights=None, early_stop=99999, split=None):
        """
        Optimize indicator parameters to maximize correlation
        :param X: Historical dataset
        :param y: Target used to measure correlation.  Can be a subset index of X
        :param trials: Number of optimization trials per indicator set
        :param indicators: List of indicators to optimize
        :param ranges: Parameter search space
        :param spearman: Perform spearman vs pearson correlation
        :param weights: Optional weights sharing the same index as y
        :param early_stop: Max number of optimization trials before stopping
        :param split: Index cut points defining time periods
        """

        self.fitted = []  # List containing each indicator completed study
        X.columns = X.columns.str.lower()  # columns must be lower case
        pool = ProcessPool(nodes=self.n_jobs)  # Set parallel cores

        # Package level optimization
        if 'tta' in indicators:
            indicators = indicators + talib_indicators
            indicators.remove('tta')
        if 'pta' in indicators:
            indicators = indicators + pandas_ta_indicators
            indicators.remove('pta')
        if 'fta' in indicators:
            indicators = indicators + finta_indicatrs
            indicators.remove('fta')
        if 'all' in indicators:
            indicators = talib_indicators + pandas_ta_indicators + finta_indicatrs
        indicators = list(OrderedDict.fromkeys(indicators))

        # Create textual representation of function in Optuna format
        # Example: 'tta.RSI(X.close, length=trial.suggest_int(\'timeperiod1\', 2, 1500))'
        # Utilizes the signature of the indicator (ie user parameters) if available
        # TTA uses help docstrings as signature is not available in C bindings
        # Parameters contained in config.py are tuned

        # Iterate user defined search space ranges
        for low, high in ranges:
            if low <= 1:
                raise ValueError("Range low must be > 1")
            if high >= len(X):
                raise ValueError(f"Range high:{high} must be > length of X:{len(X)}")

            # Iterate indicators per range
            for ind in indicators:

                # Index column to optimize if indicator returns dataframe
                idx = 0
                if ":" in ind:
                    idx = int(ind.split(":")[1])
                    ind = ind.split(":")[0]
                fn = f"{ind}("

                # If TTA indicator, use doc strings for lack of better way to
                # get indicator arguments (C binding)
                if ind[0:3] == "tta":
                    usage = eval(f"{ind}.__doc__").split(")")[0].split("(")[1]
                    params = re.sub('[^0-9a-zA-Z_\s]', '', usage).split()

                # Pandas-TA and FinTA both can be inspected for parameters
                else:
                    sig = inspect.signature(eval(ind))
                    params = sig.parameters.values()

                # Format function string
                suggest = False
                for param in params:
                    param = re.split(':|=', str(param))[0].strip()
                    if param == "open_":
                        param = "open"
                    if param == "real":
                        fn += f"X.close, "
                    elif param == "ohlc":
                        fn += f"X, "
                    elif param == "ohlcv":
                        fn += f"X, "
                    elif param in tune_series:
                        fn += f"X.{param}, "
                    elif param in tune_params:
                        suggest = True
                        fn += f"{param}=trial.suggest_int('{param}', {low}, {high}), "
                fn += ")"

                # Only optimize indicators that contain tunable parameters
                if suggest:
                    self.fitted.append(pool.apipe(Optimize(function=fn, n_trials=trials,
                        spearman=spearman).fit, X, y, idx=idx, verbose=self.verbose,
                        weights=weights, early_stop=early_stop, split=split), )
                else:
                    self.fitted.append(pool.apipe(Optimize(function=fn, n_trials=1,
                        spearman=spearman).fit, X, y, idx=idx, verbose=self.verbose,
                        weights=weights, early_stop=early_stop, split=split), )

        # Blocking wait to retrieve results
        self.fitted = [fit.get() for fit in self.fitted]

    def prune(self, top=2, studies=1):
        """
        Select most correlated with target, least intercorrelated
        :param top: Selects top x most correlated with target
        :param studies: From top x, keep y least intercorelated
        :return:
        """

        # Error checking
        if top > len(self.fitted) or studies > len(self.fitted):
            raise ValueError("Cannot prune because top or studies is >= tuned indicators")
            return
        if top < studies:
            raise ValueError(f"top {top} must be >= studies {studies}")

        # Create fitness array that maps to the correlation of each indicator study
        fitness = []
        for t in self.fitted:
            if t.split is None:
                fitness.append(t.study.best_trial.value)
            else:
                fitness.append(sum(t.study.trials[t.study.top_trial].values))
        fitness = np.array(fitness)

        # Select top x indices with most correlation to target
        fitness = fitness.argsort()[::-1][:top]  # Get sorted fitness indices of HOF

        # Gets best trial feature of each study in HOF
        features = []
        top_studies = [self.fitted[i] for i in fitness]  # Get fitness mapped studies
        for study in top_studies:
            features.append(study.res_y) # Get indicator values stored from optimization
        features = np.array(features)  # Features of HOF studies / actual indicator results

        # Correlation of HOF features
        # Create correlation table of features
        eval = np.apply_along_axis(rankdata, 1, features)
        with np.errstate(divide='ignore', invalid='ignore'):
            correlations = np.abs(np.corrcoef(eval))
        np.fill_diagonal(correlations, 0.)

        # Iteratively removes least fit individual of most correlated pairs of studies
        # IOW, finds most correlated pairs, removes lest correlated to target until x studies
        components = list(range(top))
        indices = list(range(top))
        while len(components) > studies:
            most_correlated = np.unravel_index(np.argmax(correlations), correlations.shape)
            worst = max(most_correlated)
            components.pop(worst)
            indices.remove(worst)
            correlations = correlations[:, indices][indices, :]
            indices = list(range(len(components)))

        # Save only fitted studies (overwriting all studies)
        self.fitted = [self.fitted[i] for i in fitness[components]]

    def transform(self, X, columns=None):
        """
        Given X, create features of fitted studies
        :param X: Dataset with features used to create fitted studies
        :return:
        """

        # Remove trailing identifier in column list if present
        if columns is not None:
            columns = [re.sub(r'_[0-9]+$', '', s) for s in columns]

        X.columns = X.columns.str.lower()  # columns must be lower case
        pool = ProcessPool(nodes=self.n_jobs)  # Number of jobs
        self.result = []

        # Iterate fitted studies and calculate TA with fitted parameter set
        for ind in self.fitted:
            # Create field if no columns or is in columns list
            if columns is None or ind.res_y.name in columns:
                self.result.append(pool.apipe(ind.transform, X))

        # Blocking wait for asynchronous results
        self.result = [res.get() for res in self.result]

        # Combine results into dataframe to return
        res = pd.concat(self.result, axis=1)
        return res

    def report(self, target_corr=True, features_corr=True):
        fns = []  # Function names
        cor = []  # Target Correlation
        moc = []  # Multi-Time Period Correlation
        mean_moc = []
        std_moc = []  # Multi STD
        features = []
        for fit in self.fitted:
            if fit.split is None:
                fns.append(col_name(fit.function, fit.study.best_params))
            else:
                fns.append(col_name(fit.function, fit.study.top_params))
                moc.append(fit.study.trials[fit.study.top_trial].values)
                mean_moc.append(np.mean(fit.study.trials[fit.study.top_trial].values))
                std_moc.append(np.std(fit.study.trials[fit.study.top_trial].values))

            cor.append(round(fit.res_y_corr, 6))
            features.append(fit.res_y)

        if fit.split is None:
            fitness = pd.DataFrame(cor, index=fns, columns=['Correlation']).sort_values(by=['Correlation'], ascending=False)
        else:
            fitness = pd.DataFrame(zip(cor, mean_moc, std_moc, moc), index=fns, columns=['Correlation', 'Split Mean', 'Split STD', 'Split Correlation']).sort_values(by=['Correlation'], ascending=False)

        if target_corr:
            print("\nTarget Correlation:\n")
            print(tabulate(fitness, headers=fitness.columns, tablefmt="simple"))

        eval = np.apply_along_axis(rankdata, 1, features)
        with np.errstate(divide='ignore', invalid='ignore'):
            correlations = np.abs(np.corrcoef(eval))
        correlations = pd.DataFrame(correlations, columns=fns, index=fns)
        if features_corr:
            print("\nFeature Correlation:\n")
            print(tabulate(correlations, headers=correlations.columns, tablefmt="simple"))

    def fit_times(self):
        times = [fit.time for fit in self.fitted]
        inds = [fit.function.split('(')[0] for fit in self.fitted]
        df = pd.DataFrame({'Indicator': inds, 'Times': times}).sort_values(by='Times', ascending=False)
        print(tabulate(df, headers=df.columns, tablefmt="simple"))



# from tuneta.optimize import _weighted_spearman
import pandas as pd
from tabulate import tabulate
import numpy as np


# def corr_xy(X, y):
#     X = X.loc[:, ~X.columns.duplicated()]  # Remove duplicate columns
#     fitness = []
#     for col in X.columns:
#         fitness.append(_weighted_spearman(y, X[col]))
#     df = pd.DataFrame([X.columns, fitness]).T
#     df.columns = ['Features', 'Correlation']
#     df = df.sort_values(by=['Correlation'], ascending=False)
#     print(tabulate(df, headers=df.columns, tablefmt="simple"))
#     return df


def num_nas(s, cnt=0):
    for v in s:
        if np.isnan(v):
            cnt = cnt + 1
        else:
            break
    return cnt


# Series data
tune_series = ['open', 'high', 'low', 'close', 'volume']

# Parameters to tune
tune_params = ['width', 'fast', 'slow', 'signal', 'length', 'max_lookback', 'min_lookback', 'initial',
               'lower_length', 'upper_length', 'lookback', 'medium', 'slow_w', 'roc1', 'roc2', 'roc3', 'roc4',
               'sma1', 'sma2', 'sma3', 'sma4', 'scalar', 'long', 'atr_length', 'high_length',
               'low_length', 'kc_length', 'bb_length', 'mom_length', 'swma_length', 'rvi_length', 'period',
               'er', 'ema_fast', 'ema_slow', 'period_fast', 'period_slow', 'rsi_period', 'wma_period', 'atr_period',
               'upper_period', 'lower_period', 'stoch_period', 'slow_period', 'fast_period', 'r1', 'r2', 'r3', 'r4',
               'short', 'signal', 'short_period', 'long_period', 'channel_lenght', 'channel_length',
               'average_lenght', 'average_length', 'tenkan_period', 'kijun_period', 'senkou_period', 'chikou_period',
               'k_period', 'timeperiod', 'timeperiod1', 'timeperiod2', 'timeperiod3', 'slowlimit', 'fastlimit',
               'minperiod', 'maxperiod', 'acceleration', 'accelerationlong', 'accelerationshort', 'signalperiod',
               'slowperiod', 'signalperiod', 'fastperiod', 'fastk_period', 'slowk_period', 'slowd_period',
               'fastd_period']

talib_indicators = ['tta.BBANDS', 'tta.DEMA', 'tta.EMA', 'tta.HT_TRENDLINE', 'tta.KAMA', 'tta.MA',
                        'tta.MIDPOINT', 'tta.MIDPRICE', 'tta.SAR', 'tta.SAREXT', 'tta.SMA', 'tta.T3', 'tta.TEMA',
                        'tta.TRIMA',
                        'tta.WMA', 'tta.ADX', 'tta.ADXR', 'tta.APO', 'tta.AROON:1', 'tta.AROONOSC', 'tta.BOP',
                        'tta.CCI', 'tta.CMO',
                        'tta.DX', 'tta.MACD', 'tta.MACDEXT', 'tta.MACDFIX', 'tta.MFI', 'tta.MINUS_DI', 'tta.MINUS_DM',
                        'tta.MOM',
                        'tta.PLUS_DI', 'tta.PLUS_DM', 'tta.PPO', 'tta.ROC', 'tta.ROCP', 'tta.ROCR', 'tta.ROCR100',
                        'tta.RSI', 'tta.STOCH',
                        'tta.STOCHF', 'tta.STOCHRSI', 'tta.TRIX', 'tta.ULTOSC', 'tta.WILLR', 'tta.AD', 'tta.ADOSC',
                        'tta.OBV',
                        'tta.HT_DCPERIOD', 'tta.HT_DCPHASE', 'tta.HT_PHASOR', 'tta.HT_SINE', 'tta.HT_TRENDMODE',
                        'tta.AVGPRICE', 'tta.MEDPRICE',
                        'tta.TYPPRICE', 'tta.WCLPRICE', 'tta.ATR', 'tta.NATR', 'tta.TRANGE', 'tta.CDL2CROWS',
                        'tta.CDL3BLACKCROWS',
                        'tta.CDL3INSIDE', 'tta.CDL3LINESTRIKE', 'tta.CDL3OUTSIDE', 'tta.CDL3STARSINSOUTH',
                        'tta.CDL3WHITESOLDIERS',
                        'tta.CDLABANDONEDBABY', 'tta.CDLADVANCEBLOCK', 'tta.CDLBELTHOLD', 'tta.CDLBREAKAWAY',
                        'tta.CDLCLOSINGMARUBOZU',
                        'tta.CDLCONCEALBABYSWALL', 'tta.CDLCOUNTERATTACK', 'tta.CDLDARKCLOUDCOVER', 'tta.CDLDOJI',
                        'tta.CDLDOJISTAR',
                        'tta.CDLDRAGONFLYDOJI', 'tta.CDLENGULFING', 'tta.CDLEVENINGDOJISTAR', 'tta.CDLEVENINGSTAR',
                        'tta.CDLGAPSIDESIDEWHITE',
                        'tta.CDLGRAVESTONEDOJI', 'tta.CDLHAMMER', 'tta.CDLHANGINGMAN', 'tta.CDLHARAMI',
                        'tta.CDLHARAMICROSS',
                        'tta.CDLHIGHWAVE', 'tta.CDLHIKKAKE', 'tta.CDLHIKKAKEMOD', 'tta.CDLHOMINGPIGEON',
                        'tta.CDLIDENTICAL3CROWS',
                        'tta.CDLINNECK', 'tta.CDLINVERTEDHAMMER', 'tta.CDLKICKING', 'tta.CDLKICKINGBYLENGTH',
                        'tta.CDLLADDERBOTTOM',
                        'tta.CDLLONGLEGGEDDOJI', 'tta.CDLLONGLINE', 'tta.CDLMARUBOZU', 'tta.CDLMATCHINGLOW',
                        'tta.CDLMATHOLD',
                        'tta.CDLMORNINGDOJISTAR', 'tta.CDLMORNINGSTAR', 'tta.CDLONNECK', 'tta.CDLPIERCING',
                        'tta.CDLRICKSHAWMAN',
                        'tta.CDLRISEFALL3METHODS', 'tta.CDLSEPARATINGLINES', 'tta.CDLSHOOTINGSTAR', 'tta.CDLSHORTLINE',
                        'tta.CDLSPINNINGTOP',
                        'tta.CDLSTALLEDPATTERN', 'tta.CDLSTICKSANDWICH', 'tta.CDLTAKURI', 'tta.CDLTASUKIGAP',
                        'tta.CDLTHRUSTING', 'tta.CDLTRISTAR',
                        'tta.CDLUNIQUE3RIVER', 'tta.CDLUPSIDEGAP2CROWS', 'tta.CDLXSIDEGAP3METHODS', 'tta.LINEARREG',
                        'tta.LINEARREG_ANGLE', 'tta.LINEARREG_INTERCEPT', 'tta.LINEARREG_SLOPE', 'tta.STDDEV',
                        'tta.TSF', 'tta.VAR',]

pandas_ta_indicators = ['pta.cdl_doji', 'pta.cdl_inside', 'pta.ha', 'pta.ao', 'pta.apo',
                        'pta.bias', 'pta.bop', 'pta.brar', 'pta.cci', 'pta.cfo', 'pta.cg', 'pta.cmo', 'pta.coppock',
                        'pta.er', 'pta.eri',
                        'pta.fisher', 'pta.inertia', 'pta.kdj', 'pta.kst', 'pta.macd', 'pta.mom', 'pta.pgo', 'pta.ppo',
                        'pta.psl', 'pta.pvo', 'pta.qqe', 'pta.roc', 'pta.rsi', 'pta.rvgi', 'pta.slope', 'pta.smi',
                        'pta.squeeze',
                        'pta.stoch', 'pta.stochrsi', 'pta.trix', 'pta.tsi', 'pta.uo', 'pta.willr', 'pta.dema',
                        'pta.ema', 'pta.hilo', 'pta.hl2', 'pta.hlc3', 'pta.hma', 'pta.kama', 'pta.linreg',
                        'pta.midpoint',
                        'pta.midprice', 'pta.ohlc4', 'pta.pwma', 'pta.rma', 'pta.sinwma', 'pta.sma', 'pta.ssf',
                        'pta.supertrend', 'pta.swma', 'pta.t3', 'pta.tema', 'pta.trima', 'pta.vidya', 'pta.vwap',
                        'pta.vwma',
                        'pta.wcp', 'pta.wma', 'pta.zlma', 'pta.entropy', 'pta.kurtosis', 'pta.mad', 'pta.median',
                        'pta.quantile', 'pta.skew', 'pta.stdev', 'pta.variance', 'pta.zscore', 'pta.adx', 'pta.amat',
                        'pta.aroon', 'pta.chop', 'pta.cksp', 'pta.decay', 'pta.decreasing', 'pta.increasing',
                        'pta.psar', 'pta.qstick', 'pta.ttm_trend', 'pta.vortex', 'pta.aberration',
                        'pta.accbands', 'pta.atr', 'pta.bbands', 'pta.donchian', 'pta.kc', 'pta.massi', 'pta.natr',
                        'pta.pdist', 'pta.rvi', 'pta.thermo', 'pta.true_range', 'pta.ui', 'pta.ad', 'pta.adosc',
                        'pta.aobv', 'pta.cmf', 'pta.efi', 'pta.eom', 'pta.mfi', 'pta.nvi', 'pta.obv', 'pta.pvi',
                        'pta.pvol', 'pta.pvt']

finta_indicatrs = ['fta.SMA', 'fta.SMM', 'fta.SSMA', 'fta.EMA', 'fta.DEMA', 'fta.TEMA', 'fta.TRIMA', 'fta.TRIX',
                   'fta.VAMA', 'fta.ER', 'fta.KAMA', 'fta.ZLEMA', 'fta.WMA', 'fta.HMA', 'fta.EVWMA', 'fta.VWAP',
                   'fta.SMMA', 'fta.MACD', 'fta.PPO', 'fta.VW_MACD', 'fta.EV_MACD', 'fta.MOM', 'fta.ROC', 'fta.RSI',
                   'fta.IFT_RSI', 'fta.TR', 'fta.ATR', 'fta.SAR', 'fta.BBANDS', 'fta.BBWIDTH', 'fta.MOBO',
                   'fta.PERCENT_B', 'fta.KC', 'fta.DO', 'fta.DMI', 'fta.ADX', 'fta.PIVOT', 'fta.PIVOT_FIB',
                   'fta.STOCH', 'fta.STOCHD', 'fta.STOCHRSI', 'fta.WILLIAMS', 'fta.UO', 'fta.AO', 'fta.MI',
                   'fta.VORTEX', 'fta.KST', 'fta.TSI', 'fta.TP', 'fta.ADL', 'fta.CHAIKIN', 'fta.MFI', 'fta.OBV',
                   'fta.WOBV', 'fta.VZO', 'fta.PZO', 'fta.EFI', 'fta.CFI', 'fta.EBBP', 'fta.EMV', 'fta.CCI',
                   'fta.COPP', 'fta.BASP', 'fta.BASPN', 'fta.CMO', 'fta.CHANDELIER', 'fta.WTO', 'fta.FISH', 'fta.APZ',
                   'fta.SQZMI', 'fta.VPT', 'fta.FVE', 'fta.VFI', 'fta.MSD', 'fta.STC']






import optuna
import pandas as pd
import numpy as np
from scipy.stats import rankdata
import pandas_ta as pta
from finta import TA as fta
import talib as tta
import re
import warnings
import pareto
warnings.filterwarnings("ignore")
from timeit import default_timer as timer


def col_name(function, study_best_params):
    """
    Create consistent column names given string function and params
    :param function:  Function represented as string
    :param study_best_params:  Params for function
    :return:
    """

    # Optuna string of indicator
    function_name = function.split("(")[0]

    # Optuna string of parameters
    params = re.sub('[^0-9a-zA-Z_:,]', '', str(study_best_params)).replace(",", "_")

    # Concatenate name and params to define
    col = f"{function_name}_{params}"
    return col


def _weighted_pearson(y, y_pred, w=None, pearson=True):
    """Calculate the weighted Pearson correlation coefficient."""
    if pearson:
        if w is None:
            w = np.ones(len(y))
        # idx = ~np.logical_or(np.isnan(y_pred), np.isnan(y))  # Drop NAs w/boolean mask
        # y = np.compress(idx, np.array(y))
        # y_pred = np.compress(idx, np.array(y_pred))
        # w = np.compress(idx, w)
    with np.errstate(divide='ignore', invalid='ignore'):
        y_pred_demean = y_pred - np.average(y_pred, weights=w)
        y_demean = y - np.average(y, weights=w)
        corr = ((np.sum(w * y_pred_demean * y_demean) / np.sum(w)) /
                np.sqrt((np.sum(w * y_pred_demean ** 2) *
                         np.sum(w * y_demean ** 2)) /
                        (np.sum(w) ** 2)))
    if np.isfinite(corr):
        return np.abs(corr)
    return 0.


def _weighted_spearman(y, y_pred, w=None):
    """Calculate the weighted Spearman correlation coefficient."""
    # idx = ~np.logical_or(np.isnan(y_pred), np.isnan(y))  # Drop NAs w/boolean mask
    # y = np.compress(idx, np.array(y))
    # y_pred = np.compress(idx, np.array(y_pred))
    # w = np.compress(idx, w)
    y_pred_ranked = np.apply_along_axis(rankdata, 0, y_pred)
    y_ranked = np.apply_along_axis(rankdata, 0, y)
    return _weighted_pearson(y_pred_ranked, y_ranked, w, pearson=False)


def _trial(self, trial, X):
    """
    Calculate indicator using best fitted trial over X
    :param self:  Optuna study
    :param trial:  Optuna trial
    :param X:  dataset
    :return:
    """

    # Evaluate TA defined as optuna trial string
    res = eval(self.function)

    # If return is tuple, convert to DF
    if isinstance(res, tuple):
        res = pd.DataFrame(res).T

    # Index result with X index
    res = pd.DataFrame(res, index=X.index)

    # Create consistent column names with function string and params
    name = col_name(self.function, trial.params)

    # Append integer identifier to DF with multiple columns
    if len(res.columns) > 1:
        res.columns = [f"{name}_{i}" for i in range(len(res.columns))]
    else:
        res.columns = [f"{name}"]
    return res


# Minimize difference, Maximize Total
def _min_max(study):
    """
    Multi-objective function to find best trial index with minimum deviation and max correlation
    :param study: Optuna study
    :return:
    """

    # Iterate pareto-front trials storing mean correlation and std dev
    df = []
    for trial in study.best_trials:
        df.append([trial.number, np.mean(trial.values), np.std(trial.values)])

    # Sort dataframe ascending by mean correlation
    df = pd.DataFrame(df).sort_values(by=2, ascending=True)

    # Sort df with best trial in first row
    if len(df) > 1 and len(df.iloc[:, 1:3].drop_duplicates()) > 1:

        # Create second pareto to maximize correlation and minimize stddev
        # Epsilons define precision, ie dominance over other candidates
        # Dominance is defined as x percent of stddev of stddev
        try:
            nd = pareto.eps_sort([list(df.itertuples(False))], objectives=[1, 2],
                epsilons=[1e-09, np.std(df[1])*.5], maximize=[1])
        except:
            # Something went wrong, return df
            nd = df

        # Sort remaining candidates
        nd = pd.DataFrame(nd).sort_values(by=2, ascending=True)

    # Only 1st trial so return it
    else:
        nd = df

    # Return "best" trial index
    return nd.iloc[0, 0]


def _multi_early_stopping_opt(study, trial):
    """
    Callback for to stop Optuna early with improvement
    :param study: Optuna study
    :param trial: Optuna trial
    :return:
    """

    # Get index of this trial
    this_trial = trial.number

    # Function to find "best" trial
    # Returns index of best trial (always 0 initially)
    best_trial = _min_max(study)

    # If this trial is best trial then
    # store trial index and params
    # reset early stop counter
    if this_trial == best_trial:
        study.top_trial = this_trial
        study.top_params = study.trials[this_trial].params
        study.top_value = np.mean(study.trials[this_trial].values)
        study.early_stop_count = 0

    # If best_trial (index of trials) is not the best then
    # stop if stop_count great than user defined early_stop
    # else increment count
    else:
        if study.early_stop_count > study.early_stop:
            study.early_stop_count = 0
            raise optuna.exceptions.OptunaError
        else:
            study.early_stop_count = study.early_stop_count + 1
    return


def _single_early_stopping_opt(study, trial):
    """
    Callback function to stop optuna trials early for single objective
    :param study:  Optuna study
    :param trial:  Optuna trial
    :return:
    """

    # Record first as best score
    if study.best_score is None:
        study.best_score = study.best_value

    # Record better value and reset count
    if study.best_value > study.best_score:
        study.best_score = study.best_value
        study.early_stop_count = 0

    # If count greater than user defined stop, raise error to stop
    else:
        if study.early_stop_count > study.early_stop:
            study.early_stop_count = 0
            raise optuna.exceptions.OptunaError
        else:
            study.early_stop_count = study.early_stop_count+1
    return


def _objective(self, trial, X, y, weights=None, split=None):
    """
    Objective function used in Optuna trials
    :param self:  Optuna study
    :param trial:  Optuna trial
    :param X: Entire dataset
    :param y: Target
    :param weights: Optional weights
    :param split: Split cut points
    :return:
    """

    # Generate even weights if none
    if weights is None:
        weights = pd.Series(np.ones(len(y)), index=y.index)

    # Execute trial function
    try:
        res = eval(self.function)
    except:
        raise RuntimeError(f"Optuna execution error: {self.function}")

    # If indicator result is tuple, select the one of interest
    if isinstance(res, tuple):
        res = res[self.idx]

    # Ensure result is a dataframe with same index as X
    res = pd.DataFrame(res, index=X.index)

    # If indicator result is dataframe, select the one of interest
    if len(res.columns) > 1:
        res = pd.DataFrame(res.iloc[:, self.idx])

    # y may be a subset of X, so reduce result to y and convert to series
    res_y = res.reindex(y.index).iloc[:, 0].replace([np.inf, -np.inf], np.nan)

    # Save all trial results for pruning and reporting
    # Only the best trial will eventually be saved to limit storage requirements
    self.res_y.append(res_y)  # Save results

    # Indicator result may be all NANs based on parameter set
    # Return FALSE and alert
    if np.isnan(res_y).sum() / len(res_y) > .95:  # Most or all NANs
        self.res_y_corr.append(np.zeros(len(y)))
        if split is not None:
            return tuple([False] * (len(split) - 1))
        else:
            return False

    # Obtain correlation for entire dataset
    if self.spearman:
        corr = _weighted_spearman(np.array(y), np.array(res_y), np.array(weights))
    else:
        corr = _weighted_pearson(np.array(y), np.array(res_y), np.array(weights))

    # Save correlation for res_y
    self.res_y_corr.append(corr)

    # Multi-objective optimization
    # Obtain correlation to target for each split for Optuna to maximize
    if split is not None:
        mo = []
        for i, e in enumerate(split):
            if i == 0:
                s = e
                continue

            # y could be a subset of X, use index of X to filter y
            idx = X[s:e].index

            # Filter y based on X split
            y_se = y.reindex(idx, fill_value=12345.12345)
            y_se = np.array(y_se[y_se != 12345.12345]).astype('float64')

            # Filter y predictions based on X split
            res_y_se = res_y.reindex(idx, fill_value=12345.12345)
            res_y_se = np.array(res_y_se[res_y_se != 12345.12345]).astype('float64')

            # Filter weights based on X split
            weights_se = weights.reindex(idx, fill_value=12345.12345)
            weights_se = np.array(weights_se[weights_se != 12345.12345]).astype('float64')

            # Too man NANs in split
            # try:
                # if np.isnan(res_y_se).sum() / len(res_y_se) > .95:
                    # return tuple([False] * (len(split) - 1))
            # except:
                # return tuple([False]*(len(split)-1))

            if np.isnan(res_y_se).sum() / len(res_y_se) > .95:
                return tuple([False]*(len(split)-1))

            if self.spearman:
                mo.append(_weighted_spearman(y_se, res_y_se, weights_se))
            else:
                mo.append(_weighted_pearson(y_se, res_y_se, weights_se))
            s = e
        return tuple(mo)

    # Single objective optimization return corr for entire dataset
    else:
        return corr


class Optimize():
    def __init__(self, function, n_trials=100, spearman=True):
        self.function = function
        self.n_trials = n_trials
        self.res_y = []
        self.res_y_corr = []
        self.spearman = spearman

    def fit(self, X, y, weights=None, idx=0, verbose=False, early_stop=None, split=None):
        """
        Optimize a technical indicator
        :param X: Historical dataset
        :param y: Target
        :param weights: Optional weights
        :param idx: Column to maximize if TA returnes multiple
        :param verbose: Verbosity level
        :param early_stop: Number of trials to perform without improvement before stopping
        :param split: Split points for multi-objective optimization
        :return:
        """

        start_time = timer()
        self.idx = idx
        self.split = split

        # Display Optuna trial messages
        if not verbose:
            optuna.logging.set_verbosity(optuna.logging.ERROR)

        # Single objective optimization
        if split is None:

            # Create optuna study maximizing correlation
            self.study = optuna.create_study(direction='maximize', study_name=self.function)

            # Set required early stopping variables
            self.study.early_stop = early_stop
            self.study.early_stop_count = 0
            self.study.best_score = None

            # Start optimization trial
            try:
                self.study.optimize(lambda trial: _objective(self, trial, X, y, weights),
                    n_trials=self.n_trials, callbacks=[_single_early_stopping_opt])

            # Early stopping (not officially supported by Optuna)
            except optuna.exceptions.OptunaError:
                pass

            # Keep only results of best trial for prune and reporting
            self.res_y = self.res_y[self.study.best_trial.number]
            self.res_y.name = col_name(self.function, self.study.best_trial.params)
            self.res_y_corr = self.res_y_corr[self.study.best_trial.number]

        # Multi objective optimization
        else:

            # Create study to maximize eash split
            sampler = optuna.samplers.NSGAIISampler()
            self.study = optuna.create_study(directions=(len(split)-1) * ['maximize'], sampler=sampler,
                                             study_name=self.function)

            # Early stopping variables
            self.study.early_stop = early_stop
            self.study.early_stop_count = 0

            # Custom best trial variables (Optuna "best" are immutable)
            self.study.top_trial = 0
            self.study.top_params = None
            self.study.top_value = None

            # Start optimization trial
            try:
                self.study.optimize(lambda trial: _objective(self, trial, X, y, weights, split),
                    n_trials=self.n_trials, callbacks=[_multi_early_stopping_opt])

            # Early stopping (not officially supported by Optuna)
            except optuna.exceptions.OptunaError:
                pass

            # Validation checks
            if np.mean(self.study.trials[self.study.top_trial].values) != self.study.top_value:
                raise RuntimeError("Top trial score invalid")
            if len(self.res_y) != len(self.study.trials) or len(self.res_y) != len(self.res_y_corr):
                raise RuntimeError("Total results does not equal trials")

            # Keep only results of best trial for prune and reporting
            self.res_y = self.res_y[self.study.top_trial]
            self.res_y.name = col_name(self.function, self.study.trials[self.study.top_trial].params)
            self.res_y_corr = self.res_y_corr[self.study.top_trial]

        end_time = timer()
        self.time = round(end_time - start_time, 2)

        return self

    def transform(self, X):
        """
        Calculate TA indicator using fittted best trial
        :param X: Datset
        :return:
        """

        # Calculate and store in features, replacing any potential non-finites (not sure needed)
        if self.split is None:
            features = _trial(self, self.study.best_trial, X)
        else:
            features = _trial(self, self.study.trials[self.study.top_trial], X)
        return features



from setuptools import setup, find_packages

try:
    requires = list()
    f = open("requirements.txt", "rb")
    for line in f.read().decode("utf-8").split("\n"):
        line = line.strip()
        if "#" in line:
            line = line[: line.find("#")].strip()
        if line:
            requires.append(line)
except:
    print("'requirements.txt' not found!")
    requires = list()

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="tuneta",
    version="0.1.23",
    author="John Richardson",
    author_email="jmrichardson@gmail.com",
    description="Optimize financial technical indicators for machine learning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jmrichardson/tuneta",
    keywords="technical analysis optimize tune indicators machine learning optuna tune hyperparameter optimization AI correlation",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Natural Language :: English",
        "Operating System :: OS Independent",
    ],
    install_requires=requires,
    python_requires='>=3.6',
    platform=["any"],
)